package org.artifactory.release.bundle;

import lombok.NonNull;
import org.artifactory.api.rest.distribution.bundle.models.ArtifactsBundleModel;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.bundle.BundleTransactionStatus;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.spring.InternalContextHelper;
import org.artifactory.storage.db.bundle.dao.ArtifactBundlesDao;
import org.artifactory.storage.db.bundle.dao.BundleBlobsDao;
import org.artifactory.storage.db.bundle.model.BundleNode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.sql.SQLException;
import java.util.List;
import java.util.Set;

import static org.artifactory.repo.system.SystemRepos.IN_TRANSIT_REPO_KEY;

/**
 * @author Inbar Tal
 */
@Service
public class InternalReleaseBundleServiceImpl implements InternalReleaseBundleService {
    private static final Logger log = LoggerFactory.getLogger(InternalReleaseBundleServiceImpl.class);
    private ArtifactBundlesDao bundlesDao;
    private BundleBlobsDao blobsDao;
    private InternalRepositoryService repoService;
    private AuthorizationService authService;
    ReleaseBundleDeleteTemplate releaseBundleDeleteTemplate;

    @Autowired
    public InternalReleaseBundleServiceImpl(
            @NonNull ArtifactBundlesDao bundlesDao,
            @NonNull BundleBlobsDao blobsDao,
            @NonNull InternalRepositoryService repoService,
            @NonNull AuthorizationService authService,
            @NonNull ReleaseBundleDeleteTemplate releaseBundleDeleteTemplate) {
        this.bundlesDao = bundlesDao;
        this.blobsDao = blobsDao;
        this.repoService = repoService;
        this.authService = authService;
        this.releaseBundleDeleteTemplate = releaseBundleDeleteTemplate;
    }

    @Override
    public void deleteBundleInternal(ArtifactsBundleModel bundleModel,
            List<BundleNode> bundleNodes,
            Set<Long> nodeIdsWithArtifactsToDelete) throws SQLException {
        log.debug("Deleting Release Bundle with id '{}'", bundleModel.getId());
        deleteBundleContent(bundleModel, bundleNodes, nodeIdsWithArtifactsToDelete);
        deleteExportedBundleFile(bundleModel);
        blobsDao.deleteBlob(bundleModel.getId());
        bundlesDao.deleteArtifactsBundle(bundleModel.getId());
    }

    private void deleteExportedBundleFile(@NonNull ArtifactsBundleModel bundleModel) {
        RepoPath filePath = ReleaseBundleUtils.buildExportFileRepoPath(
                bundleModel.getName(),
                bundleModel.getVersion(),
                bundleModel.getStoringRepo()
        );
        if (repoService.exists(filePath)) {
            repoService.undeploy(filePath);
        }
    }

    protected void deleteBundleContent(ArtifactsBundleModel artifactsBundle, List<BundleNode> bundleNodes,
            Set<Long> nodeIdsWithArtifactsToDelete) throws SQLException {
        if (BundleTransactionStatus.INPROGRESS.equals(artifactsBundle.getStatus())) {
            deleteInProgressReleaseBundleContent(artifactsBundle);
        } else {
            releaseBundleDeleteTemplate.deleteBundle(artifactsBundle, bundleNodes, nodeIdsWithArtifactsToDelete);
        }
    }

    private void deleteInProgressReleaseBundleContent(ArtifactsBundleModel artifactsBundle) throws SQLException {
        log.debug("Deleting an in-progress release bundle {}", artifactsBundle.getId());
        bundlesDao.deleteBundleNodes(artifactsBundle.getId());
        InternalContextHelper.get().getSecurityService().doAsSystem(() ->
                deleteArtifactsFromIntransitRepo(artifactsBundle.getName(), artifactsBundle.getVersion()));
    }

    @Override
    public void undeployIfNeeded(ArtifactsBundleModel artifactsBundle,
            Set<Long> nodeIdsWithArtifactsToDelete, BundleNode bundleNode) {
        log.debug("Deleting repoPath: {}. bundle name: {} ", bundleNode.getRepoPath(), artifactsBundle.getName());
        if (nodeIdsWithArtifactsToDelete.contains(bundleNode.getId())) {
            repoService.undeploy(RepoPathFactory.create(bundleNode.getRepoPath()));
        } else {
            log.info("Artifact {} is related to a different release bundle and will not be deleted.",
                            bundleNode.getRepoPath());
        }
    }

    @Override
    public void deleteBundleNodesInnerTransaction(ArtifactsBundleModel artifactsBundle) throws SQLException {
        log.debug("Deleting release bundle {} on new Transaction", artifactsBundle.getId());
        bundlesDao.deleteBundleNodes(artifactsBundle.getId());
    }

    /**
     * Delete the temp folder and all of its content
     */
    private void deleteArtifactsFromIntransitRepo(String bundleName, String bundleVersion) {
        String txPath = String.join("/", IN_TRANSIT_REPO_KEY, bundleName, bundleVersion);
        repoService.undeploy(RepoPathFactory.create(txPath));
    }
}
