package org.artifactory.download;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.download.FolderDownloadConfigDescriptor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.mockito.Spy;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Tom Arie
 */
public class FolderDownloadServiceImplTest {

    // default config params
    boolean enabled = true;
    boolean enabledForAnonymous = false;
    int maxDownloadSizeMb = 1024;
    long maxFiles = 5000;
    int maxConcurrentRequests = 10;
    boolean enabledEmptyDirectories = false;

    private CentralConfigDescriptor oldCentralConfigDescriptor;

    @Mock
    private ArtifactoryContext context;

    @Mock
    private CentralConfigService centralConfigService;

    @Spy
    @InjectMocks
    private FolderDownloadServiceImpl folderDownloadService;

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.openMocks(this);
        when(context.beanForType(CentralConfigService.class)).thenReturn(centralConfigService);
        ArtifactoryContextThreadBinder.bind(context);
        FolderDownloadConfigDescriptor oldFolderDownloadConfigDescriptor =
                getFolderDownloadDescriptor(enabled, enabledForAnonymous, maxDownloadSizeMb,
                        maxFiles, maxConcurrentRequests, enabledEmptyDirectories);
        oldCentralConfigDescriptor = mockConfigDescriptor(oldFolderDownloadConfigDescriptor);
    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test(dataProvider = "provideFolderDownloadConfigValues")
    public void testReload(boolean enabled, boolean enabledForAnonymous, int maxDownloadSizeMb, long maxFiles,
            int maxConcurrentRequests, boolean enabledEmptyDirectories, int numOfSetParamsInvokes) {
        //given:
        FolderDownloadConfigDescriptor newFolderDownloadDescriptor = getFolderDownloadDescriptor(enabled,
                enabledForAnonymous, maxDownloadSizeMb, maxFiles, maxConcurrentRequests, enabledEmptyDirectories);
        FolderDownloadConfigDescriptor spyNewFolderDownloadDescriptor = Mockito.spy(newFolderDownloadDescriptor);
        CentralConfigDescriptor centralConfigDescriptor = mockConfigDescriptor(spyNewFolderDownloadDescriptor);
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);
        folderDownloadService.init();
        //when:
        folderDownloadService.reload(oldCentralConfigDescriptor, null);
        //then:
        verify(folderDownloadService, times(numOfSetParamsInvokes)).
                setFolderDownloadConfigBaseParams(spyNewFolderDownloadDescriptor);
    }

    private CentralConfigDescriptor mockConfigDescriptor(
            FolderDownloadConfigDescriptor spyNewFolderDownloadDescriptor) {
        CentralConfigDescriptor centralConfigDescriptor = Mockito.mock(CentralConfigDescriptor.class);
        when(centralConfigDescriptor.getFolderDownloadConfig()).thenReturn(spyNewFolderDownloadDescriptor);
        return centralConfigDescriptor;
    }

    private FolderDownloadConfigDescriptor getFolderDownloadDescriptor(boolean enabled, boolean enabledForAnonymous,
            int maxDownloadSizeMb, long maxFiles, int maxConcurrentRequests, boolean enabledEmptyDirectories) {
        FolderDownloadConfigDescriptor folderDownloadDescriptor = new FolderDownloadConfigDescriptor();
        folderDownloadDescriptor.setEnabled(enabled);
        folderDownloadDescriptor.setEnabledForAnonymous(enabledForAnonymous);
        folderDownloadDescriptor.setMaxDownloadSizeMb(maxDownloadSizeMb);
        folderDownloadDescriptor.setMaxFiles(maxFiles);
        folderDownloadDescriptor.setMaxConcurrentRequests(maxConcurrentRequests);
        folderDownloadDescriptor.setEnabledEmptyDirectories(enabledEmptyDirectories);
        return folderDownloadDescriptor;
    }

    @DataProvider
    public static Object[][] provideFolderDownloadConfigValues() {
        return new Object[][] {
                // latest value represents the amount of calls performed
                // to set folder download config values. if two configs are different
                // we are expecting to receive 2 calls, otherwise, 1 because setFolderDownloadConfigBaseParams
                // is invoked on both init of folderDownload config and after two configs are different
                // {true, false, 1024, 5000, 10, false, 1} - default config for testing
                // enabled, enabledForAnonymous, maxDownloadSizeMb, maxFiles,
                // maxConcurrentRequests, enabledEmptyDirectories, numOfSetParamsInvokes
                {false, false, 1024, 5000, 10, false, 2},
                {true, false, 1024, 5000, 10, true, 2},
                {true, true, 1024, 5000, 10, true, 2},
                {true, false, 1000, 5000, 10, true, 2},
                {true, false, 1024, 500, 10, true, 2},
                {true, false, 1024, 5000, 8, true, 2},
                {true, false, 1024, 5000, 10, false, 1}
        };
    }
}
