package org.artifactory.engine;

import org.apache.commons.collections.iterators.IteratorEnumeration;
import org.apache.commons.io.input.CountingInputStream;
import org.apache.http.HttpHeaders;
import org.apache.http.HttpStatus;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.SecurityService;
import org.artifactory.common.ConstantValues;
import org.artifactory.fs.RepoResource;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.security.SystemAuthenticationToken;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.traffic.TrafficService;
import org.artifactory.traffic.entry.TrafficEntry;
import org.artifactory.traffic.v2.TrafficEntryV2;
import org.jetbrains.annotations.NotNull;
import org.jfrog.storage.binstore.ifc.ArtificialInputStream;
import org.mockito.ArgumentCaptor;
import org.springframework.security.core.context.SecurityContextHolder;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.function.Function;

import static org.artifactory.api.security.access.UserTokenSpec.ARTIFACTORY_FEDERATION_USER_PREFIX;
import static org.artifactory.api.security.access.UserTokenSpec.TOKEN_PREFIX;
import static org.artifactory.repo.system.SystemRepos.LOG_SHIPPING_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.USAGE_REPO_KEY;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.fail;

public class UploadServiceTrafficLogTest extends ArtifactoryHomeBoundTest {

    ArtifactoryContext context;

    @BeforeClass
    public void init() {
        context = mock(ArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterClass
    public void unbind() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void withoutError() {
        TrafficEntryV2 captured = getTrafficEntry(false);
        assertValid(captured);
    }

    @Test
    public void withError() {
        TrafficEntryV2 captured = getTrafficEntry(true);
        assertValid(captured);
    }

    private void assertValid(TrafficEntryV2 captured) {
        assertEquals(captured.getAction(), TrafficEntryV2.TrafficAction.UPLOAD);
        assertEquals(captured.getActor(), TrafficEntryV2.TrafficActor.USER);
        assertEquals(captured.getRedirectTarget(), TrafficEntryV2.TrafficTarget.DIRECT);
        assertEquals(captured.getUserAddress(), "127.0.0.1");
        assertEquals(captured.getRepoPath(), "a:b");
        assertEquals(captured.getSize(), 42);
        assertEquals(captured.getRequestedSize(), 42);
        assertEquals(captured.getTransferredSize(), 40);
    }

    private TrafficEntryV2 getTrafficEntry(boolean simulateError) {
        prepareAuth("");
        TrafficService trafficService = mock(TrafficService.class);
        CountingInputStream countingInputStream = mock(CountingInputStream.class);
        when(countingInputStream.getByteCount()).thenReturn(40L);
        UploadServiceTestStub uploadService = new UploadServiceTestStub(trafficService,
                inputStream -> countingInputStream);
        ArtifactoryRequest request = mock(ArtifactoryRequest.class);
        when(request.getRepoKey()).thenReturn("repo");
        ArtifactoryResponse response = mock(ArtifactoryResponse.class);
        LocalRepo repo = mock(LocalRepo.class);
        RepoResource res = getRepoResource(42L);
        uploadService.simulateError = simulateError;
        try {
            uploadService.uploadItemWithProvidedContent(request, response, repo, res);
        } catch (Exception e) {
            if (!simulateError) {
                fail();
            }
        }
        return verifyAndGetTrafficEntryV2(trafficService);
    }

    private TrafficEntryV2 verifyAndGetTrafficEntryV2(TrafficService trafficService) {
        ArgumentCaptor<TrafficEntryV2> captor = ArgumentCaptor.forClass(TrafficEntryV2.class);
        verify(trafficService, times(1))
                .handleTrafficEntry(captor.capture());
        return captor.getValue();
    }

    @Test
    public void testUploadWithArtificialInputStream() throws IOException, RepoRejectException {
        prepareAuth(TOKEN_PREFIX + ARTIFACTORY_FEDERATION_USER_PREFIX + "a_to_b");
        TrafficService trafficService = mock(TrafficService.class);
        ArtificialInputStream stream = mock(ArtificialInputStream.class);
        when(stream.getLength()).thenReturn(55L);
        UploadServiceTestStub uploadService = new UploadServiceTestStub(trafficService, null);
        ArtifactoryRequest request = mock(ArtifactoryRequest.class);
        when(request.getInputStream()).thenReturn(stream);
        ArtifactoryResponse response = mock(ArtifactoryResponse.class);
        when(response.getStatus()).thenReturn(HttpStatus.SC_OK);
        LocalRepo repo = mock(LocalRepo.class);
        RepoResource res = getRepoResource(55L);
        when(request.getRepoKey()).thenReturn("repo");
        uploadService.uploadItemWithProvidedContent(request, response, repo, res);

        TrafficEntryV2 value = verifyAndGetTrafficEntryV2(trafficService);

        assertEquals(value.getTransferredSize(), 55);
        assertEquals(value.getActor(), TrafficEntryV2.TrafficActor.FEDERATION);
    }

    @Test
    public void testUploadWithLogShippingAndUserSystem() throws IOException, RepoRejectException {
        prepareAuth(SecurityService.USER_SYSTEM);
        TrafficService trafficService = mock(TrafficService.class);
        ArtificialInputStream stream = mock(ArtificialInputStream.class);
        SystemAuthenticationToken authentication = mock(SystemAuthenticationToken.class);
        SecurityContextHolder.getContext().setAuthentication(authentication);
        when(stream.getLength()).thenReturn(55L);
        UploadServiceTestStub uploadService = new UploadServiceTestStub(trafficService, null);
        ArtifactoryRequest request = mock(ArtifactoryRequest.class);
        when(request.getInputStream()).thenReturn(stream);
        when(request.getRepoKey()).thenReturn(LOG_SHIPPING_REPO_KEY);
        when(request.getHeaderValues(HttpHeaders.USER_AGENT))
                .thenReturn(new IteratorEnumeration(List.of("observability", "observability").iterator()));
        ArtifactoryResponse response = mock(ArtifactoryResponse.class);
        when(response.getStatus()).thenReturn(HttpStatus.SC_OK);
        LocalRepo repo = mock(LocalRepo.class);
        RepoResource res = getRepoResource(55L);

        uploadService.uploadItemWithProvidedContent(request, response, repo, res);

        TrafficEntryV2 value = verifyAndGetTrafficEntryV2(trafficService);

        assertEquals(value.getTransferredSize(), 55);
        assertEquals(value.getActor(), TrafficEntryV2.TrafficActor.OBSERVABILITY);
        assertEquals(value.getAction(), TrafficEntryV2.TrafficAction.UPLOAD);
        SecurityContextHolder.clearContext();
    }

    @Test
    public void testUploadWithUsageAndUserSystem() throws IOException, RepoRejectException {
        prepareAuth(SecurityService.USER_SYSTEM);
        TrafficService trafficService = mock(TrafficService.class);
        ArtificialInputStream stream = mock(ArtificialInputStream.class);
        SystemAuthenticationToken authentication = mock(SystemAuthenticationToken.class);
        SecurityContextHolder.getContext().setAuthentication(authentication);
        when(stream.getLength()).thenReturn(55L);
        UploadServiceTestStub uploadService = new UploadServiceTestStub(trafficService, null);
        ArtifactoryRequest request = mock(ArtifactoryRequest.class);
        when(request.getInputStream()).thenReturn(stream);
        when(request.getRepoKey()).thenReturn(USAGE_REPO_KEY);
        when(request.getHeaderValues(HttpHeaders.USER_AGENT))
                .thenReturn(new IteratorEnumeration(List.of("observability", "observability").iterator()));
        ArtifactoryResponse response = mock(ArtifactoryResponse.class);
        when(response.getStatus()).thenReturn(HttpStatus.SC_OK);
        LocalRepo repo = mock(LocalRepo.class);
        RepoResource res = getRepoResource(55L);

        uploadService.uploadItemWithProvidedContent(request, response, repo, res);

        TrafficEntryV2 value = verifyAndGetTrafficEntryV2(trafficService);

        assertEquals(value.getTransferredSize(), 55);
        assertEquals(value.getActor(), TrafficEntryV2.TrafficActor.OBSERVABILITY_USAGE);
        assertEquals(value.getAction(), TrafficEntryV2.TrafficAction.UPLOAD);
        SecurityContextHolder.clearContext();
    }

    @Test
    public void testUploadFailsWithArtificialInputStream() throws IOException, RepoRejectException {
        prepareAuth(TOKEN_PREFIX + ARTIFACTORY_FEDERATION_USER_PREFIX + "a_to_b");
        TrafficService trafficService = mock(TrafficService.class);
        ArtificialInputStream stream = mock(ArtificialInputStream.class);
        when(stream.getLength()).thenReturn(55L);
        UploadServiceTestStub uploadService = new UploadServiceTestStub(trafficService, null);
        ArtifactoryRequest request = mock(ArtifactoryRequest.class);
        when(request.getInputStream()).thenReturn(stream);
        ArtifactoryResponse response = mock(ArtifactoryResponse.class);
        LocalRepo repo = mock(LocalRepo.class);
        RepoResource res = getRepoResource(55L);

        uploadService.uploadItemWithProvidedContent(request, response, repo, res);
        verify(trafficService, times(0))
                .handleTrafficEntry(any(TrafficEntryV2.class));
        verify(trafficService, times(0))
                .handleTrafficEntry(any(TrafficEntry.class));
    }

    @NotNull
    private RepoResource getRepoResource(long size) {
        RepoResource res = mock(RepoResource.class);
        when(res.getSize()).thenReturn(size);
        RepoPath repoPath = new RepoPathImpl("a", "b");
        when(res.getRepoPath()).thenReturn(repoPath);
        return res;
    }

    private void prepareAuth(String userName) {
        AuthorizationService authorizationService = mock(AuthorizationService.class);
        when(authorizationService.currentUsername()).thenReturn(userName);
        when(context.beanForType(AuthorizationService.class)).thenReturn(authorizationService);
        homeStub.getArtifactoryProperties()
                .setProperty(ConstantValues.test.getPropertyName(), "true");
    }

    private static class UploadServiceTestStub extends UploadServiceImpl {

        private boolean simulateError;

        public UploadServiceTestStub(TrafficService trafficService,
                Function<InputStream, CountingInputStream> streamFactory) {
            super();
            setTrafficService(trafficService);
            setStreamFactory(streamFactory);
        }

        @Override
        void uploadItemWithContent(ArtifactoryRequest request, ArtifactoryResponse response, LocalRepo repo,
                RepoResource res, InputStream inputStream) throws IOException {
            if (simulateError) {
                throw new IOException();
            }
        }
    }
}
