package org.artifactory.security.ssh;

import org.apache.sshd.server.session.ServerSession;
import org.artifactory.api.security.UserInfoBuilder;
import org.artifactory.storage.security.service.UserGroupStoreService;
import org.jfrog.security.util.BCProviderFactory;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import java.math.BigInteger;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.RSAPublicKeySpec;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

/**
 * @author Noam Y. Tenne
 */
class PublicKeyAuthenticatorTest {

    @Test
    void AuthenticatePublicKey() throws NoSuchAlgorithmException, InvalidKeySpecException {
        UserGroupStoreService userGroupStoreService = mock(UserGroupStoreService.class);
        ServerSession serverSession = mock(ServerSession.class);
        PublicKeyAuthenticator authenticator = new PublicKeyAuthenticator(userGroupStoreService);
        KeyFactory keyFactory = KeyFactory.getInstance("RSA", BCProviderFactory.getProvider());
        PublicKey publicKey = keyFactory.generatePublic(new RSAPublicKeySpec(new BigInteger("1"), new BigInteger("3")));


        when(userGroupStoreService.findUserByProperty("sshPublicKey", "ssh-rsa AAAAB3NzaC1yc2EAAAABAwAAAAEB", true))
                .thenReturn(new UserInfoBuilder("jim").build());
        boolean result = authenticator.authenticate("git", publicKey, serverSession);

        Assertions.assertTrue(result);
        verify(userGroupStoreService, times(1))
                .findUserByProperty("sshPublicKey", "ssh-rsa AAAAB3NzaC1yc2EAAAABAwAAAAEB", true);
        verify(serverSession, times(1)).setAttribute(Mockito.any(), Mockito.eq("jim"));
        verifyNoMoreInteractions(userGroupStoreService, serverSession);
    }

    @Test
    void AuthenticatePublicKeyAndFindNoUser() throws NoSuchAlgorithmException, InvalidKeySpecException {
        UserGroupStoreService userGroupStoreService = mock(UserGroupStoreService.class);
        ServerSession serverSession = mock(ServerSession.class);
        PublicKeyAuthenticator authenticator = new PublicKeyAuthenticator(userGroupStoreService);
        KeyFactory keyFactory = KeyFactory.getInstance("RSA", BCProviderFactory.getProvider());
        PublicKey publicKey = keyFactory.generatePublic(new RSAPublicKeySpec(new BigInteger("1"), new BigInteger("3")));

        when(userGroupStoreService.findUserByProperty("sshPublicKey", "ssh-rsa AAAAB3NzaC1yc2EAAAABAwAAAAEB", true))
                .thenReturn(null);
        boolean result = authenticator.authenticate("git", publicKey, serverSession);

        Assertions.assertFalse(result);
        verify(userGroupStoreService, times(1))
                .findUserByProperty("sshPublicKey", "ssh-rsa AAAAB3NzaC1yc2EAAAABAwAAAAEB", true);
        verifyNoMoreInteractions(userGroupStoreService, serverSession);
    }
}