package org.artifactory.storage.db.binstore.service;

import com.google.common.collect.Lists;
import com.google.common.collect.MapMaker;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.storage.BinariesInfo;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.storage.binstore.service.BinariesGarbageCollectorService;
import org.artifactory.storage.binstore.service.GarbageCollectListenerResultStatus;
import org.artifactory.storage.binstore.service.GarbageCollectionDatastoreService;
import org.artifactory.storage.binstore.service.GarbageCollectorInfo;
import org.artifactory.storage.binstore.service.GarbageCollectorStrategy;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.binstore.dao.BinariesDao;
import org.artifactory.storage.db.binstore.entity.BinaryEntity;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.common.concurrency.ConflictGuard;
import org.jfrog.common.concurrency.JvmConflictsGuard;
import org.jfrog.security.util.Pair;
import org.jfrog.storage.binstore.common.BinaryElementImpl;
import org.jfrog.storage.binstore.config.model.ProviderMetaData;
import org.jfrog.storage.binstore.exceptions.SignedUrlException;
import org.jfrog.storage.binstore.ifc.BinaryStoreFederatedMember;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;
import org.jfrog.storage.binstore.ifc.model.BinaryElementRequest;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersResponseMap;
import org.jfrog.storage.binstore.ifc.provider.BinaryProvider;
import org.jfrog.storage.binstore.manager.BinaryProviderManager;
import org.jfrog.storage.binstore.providers.base.CloudBinaryProvider;
import org.jfrog.storage.binstore.providers.base.MutableBinaryProvider;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.testng.collections.Maps;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.REPOSITORY_KEY;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Shay Bagants
 * Author: Ivan Perepelytsya
 * Date: 6/3/20.
 */
@Test
public class BinaryServiceImplTest extends ArtifactoryHomeBoundTest {

    @InjectMocks
    private BinaryServiceImpl binaryService;

    @Mock
    private BinaryProviderManager binaryProviderManager;
    @Mock
    private CloudBinaryProvider cloudBinaryProvider;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private HaCommonAddon haCommonAddon;
    @Mock
    private AolService aolService;
    @Mock
    private MutableBinaryProvider mutableBinaryProvider;
    @Mock
    private ProviderMetaData providerMetaData;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @Mock
    private BinariesGarbageCollectorService gcService;
    @Mock
    private TrashService trashService;
    @Mock
    private BinariesDao binariesDao;
    @Mock
    private InternalDbService dbService;
    @Mock
    private GarbageCollectionDatastoreService garbageCollectionDatastoreService;
    @Mock
    private DeleteBlobInfosWrapper deleteBlobInfosWrapper;
    private int originalBatchSize;

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.openMocks(this);
        when(artifactoryContext.beanForType(BinariesGarbageCollectorService.class)).thenReturn(gcService);
        when(artifactoryContext.beanForType(TrashService.class)).thenReturn(trashService);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        ReflectionTestUtils.setField(binaryService, "binariesDao", binariesDao);
        ReflectionTestUtils
                .setField(binaryService, "garbageCollectionDatastoreService", garbageCollectionDatastoreService);
        binaryService.setDbService(dbService);
        binaryService.setDeleteBlobInfosWrapper(deleteBlobInfosWrapper);
        originalBatchSize = ConstantValues.trashcanMaxSearchResults.getInt();
        ArtifactoryHome.get().getArtifactoryProperties().setProperty(
                ConstantValues.trashcanMaxSearchResults.getPropertyName(), "3");
        when(artifactoryContext.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(binaryProviderManager.getNextCloudProvider()).thenReturn(cloudBinaryProvider);
        when(cloudBinaryProvider.getMutableBinaryProvider()).thenReturn(mutableBinaryProvider);
        when(mutableBinaryProvider.getProviderMetaData()).thenReturn(providerMetaData);
        when(artifactoryContext.beanForType(AolService.class)).thenReturn(aolService);
        when(addonsManager.addonByType(HaCommonAddon.class)).thenReturn(haCommonAddon);
        when(haCommonAddon.getConflictsGuard(NonStandardLockCategory.FORCE_BINARY_SHARDING_OPTIMIZATION.getName()))
                .thenReturn(new JvmConflictsGuard<>(1));

    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.get().getArtifactoryProperties().setProperty(
                ConstantValues.trashcanMaxSearchResults.getPropertyName(), String.valueOf(originalBatchSize));
    }

    public void testStartGarbageCollect() throws SQLException {
        doThrow(new SQLException("message")).when(binariesDao).getCountAndTotalSize();
        when(trashService.isTrashcanEnabled()).thenReturn(true);

        AtomicInteger fullGcErrorCount = new AtomicInteger(0);
        for (int i = 0; i < 100; i++) {
            try {
                binaryService.startGarbageCollect();
            } catch (Exception e) {
                if (e.getMessage().equals("Could not find potential Binaries to delete!")) {
                    fullGcErrorCount.incrementAndGet();
                }
            }
        }
        verify(gcService, times(100)).startGCByStrategy(any());
        // ensure full gc was triggered once
        Assert.assertEquals(5, fullGcErrorCount.get());
        AtomicInteger gcIteration = (AtomicInteger) ReflectionTestUtils.getField(binaryService, "gcIteration");
        Assert.assertNotNull(gcIteration);
        Assert.assertEquals(gcIteration.get(), 0);
        binaryService.startGarbageCollect();
        Assert.assertEquals(gcIteration.get(), 1);
    }

    @Test
    public void runFullGarbageCollectCompletedTest() throws SQLException, InterruptedException {
        BinariesInfo binariesInfo = mock(BinariesInfo.class);
        when(binariesInfo.getBinariesCount()).thenReturn(10L);
        when(binariesInfo.getBinariesSize()).thenReturn(200L);
        when(binariesDao.getCountAndTotalSize()).thenReturn(binariesInfo);
        when(dbService.invokeInTransaction(eq("BinaryCleaner#sha1a1"), any())).thenReturn(true);
        Collection<BinaryEntity> results = Lists.newArrayList();
        for (int i = 0; i < 10; i++) {
            BinaryEntity binaryEntity = new BinaryEntity("sha1a" + i, "sha2a" + i, "sha5a" + i, 5 + i);
            results.add(binaryEntity);
        }
        doNothing().when(deleteBlobInfosWrapper).deleteBlobInfosAsync(any());
        when(binariesDao.findPotentialDeletion()).thenReturn(results);
        GarbageCollectorInfo garbageCollectorInfo = binaryService.runFullGarbageCollect();
        assertEquals(garbageCollectorInfo.status, GarbageCollectListenerResultStatus.COMPLETED);
        assertEquals(garbageCollectorInfo.failures, 9);
        assertEquals(garbageCollectorInfo.strategy, GarbageCollectorStrategy.FULL);
        when(garbageCollectionDatastoreService.getLastGCRunInfo()).thenReturn(garbageCollectorInfo);
        verify(binaryProviderManager).optimize(false);
        JvmConflictsGuard<Object> conflictsGuard = new JvmConflictsGuard<>(1);
        when(haCommonAddon.getConflictsGuard(NonStandardLockCategory.FORCE_BINARY_SHARDING_OPTIMIZATION.getName()))
                .thenReturn(
                        conflictsGuard);
        ConflictGuard lock = conflictsGuard.getLock("forceOptimizationOnce");
        lock.tryToLock(10, TimeUnit.SECONDS);
        binaryService.runFullGarbageCollect();
        assertEquals(garbageCollectionDatastoreService.getLastGCRunInfo(), garbageCollectorInfo);
        verify(binaryProviderManager).optimize(true);
    }

    @Test
    public void runFullGarbageCollectAbortedTest() throws SQLException {
        BinariesInfo binariesInfo = mock(BinariesInfo.class);
        when(binariesInfo.getBinariesCount()).thenReturn(10L);
        when(binariesInfo.getBinariesSize()).thenReturn(200L);
        when(binariesDao.getCountAndTotalSize()).thenReturn(binariesInfo);
        Collection<BinaryEntity> results = Lists.newArrayList();
        when(dbService.invokeInTransaction(eq("BinaryCleaner#sha1a1"), any())).thenReturn(true);
        for (int i = 0; i < 20; i++) {
            BinaryEntity binaryEntity = new BinaryEntity("sha1a" + i, "sha2a" + i, "sha5a" + i, 5 + i);
            results.add(binaryEntity);
        }
        doNothing().when(deleteBlobInfosWrapper).deleteBlobInfosAsync(any());
        when(binariesDao.findPotentialDeletion()).thenReturn(results);
        GarbageCollectorInfo garbageCollectorInfo = binaryService.runFullGarbageCollect();
        assertEquals(garbageCollectorInfo.status, GarbageCollectListenerResultStatus.UNCOMPLETED);
        assertEquals(garbageCollectorInfo.failures, 10);
        assertEquals(garbageCollectorInfo.strategy, GarbageCollectorStrategy.FULL);
    }

    @Test
    public void testCdnIsAllowed() {
        when(aolService.isAol()).thenReturn(true);
        when(providerMetaData.getType()).thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        when(addonsManager.isEnterpriseLicensed()).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertTrue(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedNotAol() {
        when(aolService.isAol()).thenReturn(false);
        when(providerMetaData.getType()).thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        when(addonsManager.isEnterpriseLicensed()).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedNotS3() {
        when(providerMetaData.getType()).thenReturn("GCP");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        when(addonsManager.isEnterpriseLicensed()).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedNoRedirects() {
        when(providerMetaData.getType()).thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(false);
        when(addonsManager.isEnterpriseLicensed()).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedWrongLicense() {
        when(providerMetaData.getType()).thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testGetSignedUrl() {
        // setup
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 100);
        CloudBinaryProvider mockedProvider = mock(CloudBinaryProvider.class);
        when(binaryProviderManager.getNextCloudProvider()).thenReturn(mockedProvider);
        when(mockedProvider.getSignedUrl(binaryElement)).thenReturn("http://signed.url");

        // test
        String signedUrl = binaryService.getSignedUrl(binaryElement);

        // verify
        assertEquals(signedUrl, "http://signed.url");
    }

    @Test(expectedExceptions = SignedUrlException.class)
    public void testGetSignedUrl_NoCloudProvider() {
        // setup
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 100);
        when(binaryProviderManager.getNextCloudProvider()).thenReturn(null);
        // test
        binaryService.getSignedUrl(binaryElement);
    }

    @Test(expectedExceptions = SignedUrlException.class)
    public void testGetSignedUrl_EmptyUrl() {
        // setup
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 100);
        CloudBinaryProvider mockedProvider = mock(CloudBinaryProvider.class);
        when(binaryProviderManager.getNextCloudProvider()).thenReturn(mockedProvider);
        when(mockedProvider.getSignedUrl(binaryElement)).thenReturn(null);

        // test
        binaryService.getSignedUrl(binaryElement);
    }

    @Test
    public void testUpdateFederatedRepoBinaryProvider() {
        List<BinaryStoreFederatedMember> addMembers = new ArrayList<>();
        addMembers.add(new BinaryStoreFederatedMember("localKey", "remoteKey", "remoteUrl", null, "token"));
        List<BinaryStoreFederatedMember> removedMembers = new ArrayList<>();
        removedMembers
                .add(new BinaryStoreFederatedMember("removedKey", "remoteRemovedKey", "remoteUrl", null, "token"));

        binaryService.updateFederateRepoBinaryProvider(addMembers, removedMembers);
        verify(binaryProviderManager, times(1)).signalMirrorChanged(addMembers, removedMembers);
    }

    @Test
    public void testGetBinaryForChecksumDeploy() {
        Map<String, String> headers = Maps.newHashMap();
        headers.put(REPOSITORY_KEY.getHeaderName(), "repo-key");
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 1, headers);
        BinaryProvider binaryProvider = mock(BinaryProvider.class);
        BinaryProvidersResponseMap response = mock(BinaryProvidersResponseMap.class);
        when(response.exists()).thenReturn(true);
        doAnswer(invocation -> {
            final BinaryElementRequest argument = invocation.getArgument(0, BinaryElementRequest.class);
            assertEquals(argument.getHeaders(), headers);
            return response;
        }).when(binaryProvider).exists(any());

        when(binaryProviderManager.getFirstBinaryProvider()).thenReturn(binaryProvider);
        ConcurrentMap<String, Pair<AtomicInteger, Long>> deleteProtectedBinaries = new MapMaker().makeMap();
        ReflectionTestUtils.setField(binaryService, "deleteProtectedBinaries", deleteProtectedBinaries);

        binaryService.getBinaryForChecksumDeploy(binaryElement);
        verify(binaryProvider, times(1)).exists(any());
    }
}