/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.builds.buildsinfo.tabs.permissions;

import com.google.common.collect.Maps;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.security.AclService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.repo.RepoPath;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.security.BuildAcl;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.tabs.permission.PrincipalEffectivePermissions;
import org.artifactory.ui.rest.model.builds.permissions.BuildPermissionsModel;
import org.artifactory.ui.rest.service.common.EffectivePermissionHelper;
import org.artifactory.util.AolUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;

/**
 * Used in Effective Permissions tab per build, called by the Build browser
 *
 * @author Yuval Reches
 */
@Slf4j
@Component
public class GetBuildEffectivePermissionService extends GetBuildEffectivePermissionServiceBase {

    private final EffectivePermissionHelper effectivePermissionHelper;

    @Autowired
    public GetBuildEffectivePermissionService(@NonNull AuthorizationService authService,
            @NonNull UserGroupService userGroupService, BuildService buildService) {
        super(authService, userGroupService, buildService);
        this.effectivePermissionHelper = new EffectivePermissionHelper(userGroupService);
    }

    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        BuildInfo buildInfo = getAndValidateRequestDataByPathParams(request, response);
        if (buildInfo == null) {
            return;
        }
        try {
            response.iModel(
                    getPermissionModel(buildInfo.getName(), buildInfo.getNumber(), buildInfo.getStarted(),
                            buildInfo.getBuildRepo()));
        } catch (Exception e) {
            log.error("Artifactory was unable to get users and groups effective permission for build. {}",
                    e.getMessage());
            log.debug("", e);
            response.error("Artifactory was unable to get users and groups effective permissions, " +
                    "see logs for more details.");
        }
    }

    /**
     * Find and populate permission model from acl cache
     */
    public BuildPermissionsModel getPermissionModel(RepoPath repoPath) {
        List<BuildAcl> buildAcls = (repoPath.isRoot()) ? ((AclService) userGroupService).getAllBuildAcls(true) :
                userGroupService.getBuildPathAcls(repoPath, true);
        return getBuildPermissionsModel(buildAcls);
    }

    private BuildPermissionsModel getPermissionModel(String buildName, String buildNumber, String buildStarted,
            String repo) {
        List<BuildAcl> buildAcls = userGroupService.getBuildPathAcls(buildName, buildNumber, buildStarted, repo);
        return getBuildPermissionsModel(buildAcls);
    }

    private BuildPermissionsModel getBuildPermissionsModel(List<BuildAcl> buildAcls) {
        Map<String, PrincipalEffectivePermissions> userEffectivePermissions = Maps.newHashMap();
        Map<String, PrincipalEffectivePermissions> groupEffectivePermissions = Maps.newHashMap();
        boolean isAol = AolUtils.isAol();
        effectivePermissionHelper.addAdminsToMaps(userEffectivePermissions, groupEffectivePermissions, isAol);
        buildAcls.forEach(aclInfo -> effectivePermissionHelper
                .addAclInfoToAclList(aclInfo, userEffectivePermissions, groupEffectivePermissions));
        effectivePermissionHelper
                .grantGroupUsersEffectivePermissions(groupEffectivePermissions, userEffectivePermissions);
        // update response with model
        return new BuildPermissionsModel(userEffectivePermissions.values(), groupEffectivePermissions.values(),
                buildAcls);
    }

}
